import time
from datetime import datetime

import openai
from joblib import Memory
from loguru import logger
from omegaconf import DictConfig
from openai.error import AuthenticationError
from tenacity import retry, retry_if_not_exception_type, stop_after_attempt, wait_exponential

from .classes import AppUsageException

memory = Memory(".joblib_cache", verbose=0)

# TODO: try/catch! E.g.
#       APIError: HTTP code 502 from API
#       Timeout: Request timed out: HTTPSConnectionPool(host='api.openai.com', port=443): Read timed out. (read timeout=600)


def log_retry(state):
    msg = (
        f"Tenacity retry {state.fn.__name__}: {state.attempt_number=}, {state.idle_for=}, {state.seconds_since_start=}"
    )
    if state.attempt_number < 1:
        logger.warning(msg)
    else:
        logger.exception(msg)


@memory.cache()
@retry(
    wait=wait_exponential(multiplier=2, min=10, max=600),
    stop=stop_after_attempt(3),
    before_sleep=log_retry,
    retry=retry_if_not_exception_type(AppUsageException),
)
def make_call(system: str, prompt: str, llm_config: DictConfig) -> (str, int):
    if llm_config.use_localhost:
        openai.api_key = "localhost"
        openai.api_base = "http://localhost:8081"
        time.sleep(llm_config.localhost_sleep)
    else:
        openai.api_key = llm_config.api_key

    messages = [{"role": "system", "content": system}, {"role": "user", "content": prompt}]

    start = datetime.now()
    try:
        api_response = openai.ChatCompletion.create(
            model=llm_config.model, messages=messages, temperature=llm_config.temperature, top_p=llm_config.top_p
        )
    except AuthenticationError as ex:
        raise AppUsageException(str(ex)) from ex
    took = datetime.now() - start

    chat_response = api_response.choices[0].message.content
    total_tokens = int(api_response["usage"]["total_tokens"])

    logger.debug(f"{llm_config.use_localhost=}")
    logger.debug(f"{system=}")
    logger.debug(f"{prompt=}")
    logger.debug(f"{took=}")
    logger.debug("\n---- RESPONSE:")
    logger.debug(f"{chat_response=}")
    logger.debug(f"{total_tokens=}")

    return chat_response, total_tokens



import pkg_resources

package_name = "gptauthor"
version = pkg_resources.get_distribution(package_name).version

default_output_folder = "./_output/"
default_write_total_chapters = 3

# https://platform.openai.com/docs/api-reference/chat/create
default_llm_use_localhost = 0
default_llm_model = "gpt-3.5-turbo"
default_llm_temperature = 1  # Default 1 as per OpenAI docs
default_llm_top_p = 1  # Default 1 as per OpenAI docs



import re
from datetime import datetime
from pathlib import Path


def _make_safe_filename(s, max_chars=36):
    safe = s.lower()
    safe = safe.replace("title", "")
    safe = safe.replace(" ", "_")
    safe = re.sub("_+", "_", safe)
    safe = re.sub(r"[^a-zA-Z0-9_]", "", safe)
    safe = safe.strip("_")
    safe = safe.strip()
    safe = safe[:max_chars]
    return safe


def _case_insensitive_split(split, input):
    input = input.replace("**", "")  # Remove any MD bolding
    parts = re.split(split, input, flags=re.IGNORECASE)
    return parts


def gpt4_8k_price_estimate(total_tokens):
    # https://openai.com/pricing#language-models
    return (total_tokens / 1000) * 0.045  # Dec 2023


def gpt35_4k_price_estimate(total_tokens):
    # https://openai.com/pricing#language-models
    return (total_tokens / 1000) * 0.002  # Dec 2023


def synopsis_processer(synopsis_response):
    chapters_split = _case_insensitive_split("\nChapter", synopsis_response)

    title = None
    chapters = []
    for i, chapter in enumerate(chapters_split):
        if i == 0:
            title = chapter.strip()
            title = title.replace("Title:", "").replace('"', "").strip()
            title = re.sub(" +", " ", title)
        else:
            chapter_clean = chapter.replace('"', "").strip()
            chapters.append("Chapter " + chapter_clean)

    return title, chapters


def get_folder(synopsis_title: str, synopsis_chapters: list[str], llm_config: dict) -> Path:
    now = datetime.now()
    safe_title = _make_safe_filename(synopsis_title)
    num_chapters = len(synopsis_chapters)
    story_name = llm_config.story_file.split("/")[-1].split(".")[0]
    folder = (
        Path(llm_config.default_output_folder)
        / f"{story_name}"
        / f"{llm_config.model}"
        / f"{now.strftime('%Y%m%d-%H%M%S')}-v{llm_config.version}-{llm_config.model}-T{llm_config.temperature}-P{llm_config.top_p}-C{num_chapters}-{safe_title}"
    )
    folder.mkdir(parents=True, exist_ok=True)
    return folder



import re
import time
import warnings
import webbrowser
from datetime import datetime

import markdown
import typer
from loguru import logger
from rich import print
from tqdm import TqdmExperimentalWarning
from tqdm.rich import tqdm

from . import consts, llm, prompts, utils

warnings.filterwarnings("ignore", category=TqdmExperimentalWarning)


def p(message: str):
    print(message)
    logger.info(message)


def replace_chapter_text(string):
    pattern = r"\n\"?Chapter (\d+)"  # NOTE: double-quote at start is optional for match, but will leave a hanging double-quote if present
    replacement = r"\n### Chapter \1"
    new_chapter_text = re.sub(pattern, replacement, string, flags=re.IGNORECASE)
    return new_chapter_text


def user_input_continue_processing(synopsis_response_user_edited_filename):
    while True:
        print(f"This is your chance to edit the file '{synopsis_response_user_edited_filename}' before continuing.\n")
        user_input = input("Press 'C' to continue writing chapters, or 'Q' to quit: ")
        if user_input.lower() == "c":
            return True
        elif user_input.lower() == "q":
            return False
        else:
            print("Invalid input. Please try again.")


def do_writing(llm_config):
    start = time.time()
    p(
        f"Start {consts.package_name} {consts.version}, {llm_config.total_chapters=}, {llm_config.story_file=}, {llm_config.allow_user_input=}..."
    )
    p(f"Full story prompt file path: {prompts.get_yaml_file(llm_config).resolve()}")
    print()

    # ------------------------------------------------------------------------------
    # Create synopsis
    # ------------------------------------------------------------------------------
    book_description = prompts.get_book_description(llm_config)
    book_characters = prompts.get_book_characters(llm_config)
    synopsis_system = prompts.get_system("synopsis", llm_config)
    synopsis_prompt_format = prompts.get_prompt("synopsis", llm_config)
    synopsis_prompt = synopsis_prompt_format.format(
        total_chapters=str(llm_config.total_chapters),
        book_description=book_description,
        book_characters=book_characters,
    )

    synopsis_response_user_edited_filename = "synopsis_response_user_edited.txt"

    synopsis_response, synopsis_total_tokens = llm.make_call(synopsis_system, synopsis_prompt, llm_config)
    synopsis_title, synopsis_chapters = utils.synopsis_processer(synopsis_response)

    output_folder = utils.get_folder(synopsis_title, synopsis_chapters, llm_config)
    safe_llm_config = llm_config.copy()
    del safe_llm_config["api_key"]  # NOTE: remove api_key so it isn't printed
    safe_llm_config["app_version"] = consts.version
    safe_llm_config["output_folder"] = str(output_folder)
    safe_llm_config["datetime"] = datetime.now().strftime("%Y%m%d-%H%M%S")

    p(f"Writing output to folder: '{output_folder}'...")
    with open(output_folder / "_synopsis_prompt.txt", "w") as f:
        f.write(f"{synopsis_system}")
        f.write("\n---\n")
        f.write(f"{synopsis_prompt}\n")
    with open(output_folder / "synopsis_response_original.txt", "w") as f:
        f.write(f"{synopsis_response.strip()}")
    with open(output_folder / synopsis_response_user_edited_filename, "w") as f:
        f.write(f"{synopsis_response.strip()}")
    with open(output_folder / "_chapter_prompt_format.txt", "w") as f:
        f.write(prompts.get_prompt("expand-chapter-first", llm_config))
        f.write("\n---\n")
        f.write(prompts.get_prompt("expand-chapter-next", llm_config))
    with open(output_folder / "_synopsis.txt", "w") as f:
        f.write(synopsis_response)
        f.write("\n---\n")
        f.write(str(safe_llm_config))
        f.write(f"{synopsis_total_tokens=}")

    took = time.time() - start

    p("Synopsis:")
    print("```")
    p(synopsis_response)
    print("```")
    p(f"\nFinished synopsis for book '{synopsis_title}' with {len(synopsis_chapters)} chapters")
    p(f"\n{took=:.2f}s")
    p(f"Total synopsis tokens: {synopsis_total_tokens:,}")
    p(
        f"Rough synopsis GPT4 8k price: ${utils.gpt4_8k_price_estimate(synopsis_total_tokens):.2f} (estimated, check your usage!)"
    )
    p(
        f"Rough synopsis GPT3.5 4k price: ${utils.gpt35_4k_price_estimate(synopsis_total_tokens):.3f} (estimated, check your usage!)"
    )
    p(f"\n{safe_llm_config=}\n")

    if len(synopsis_title) > 100:
        logger.warning(f"Unexpected synopsis_title length! {len(synopsis_title)=}, synopsis_title='{synopsis_title}'")
        with open(output_folder / "__error.txt", "w") as f:
            f.write(f"Unexpected synopsis_title length! {len(synopsis_title)=}")
            f.write(str(safe_llm_config))
        raise typer.Exit(1)

    # ------------------------------------------------------------------------------
    # User input to continue or quit
    # ------------------------------------------------------------------------------
    if llm_config.allow_user_input and not user_input_continue_processing(
        str(output_folder / synopsis_response_user_edited_filename)
    ):
        with open(output_folder / "__aborted.txt", "w") as f:
            f.write(f"Aborted writing book: {synopsis_title}.\n\n")
            f.write(str(safe_llm_config))
        raise typer.Exit(0)

    # ------------------------------------------------------------------------------
    # Load user edited synopsis (if applicable)
    # ------------------------------------------------------------------------------
    with open(output_folder / synopsis_response_user_edited_filename, "r") as f:
        synopsis_response_original = synopsis_response
        synopsis_response = f.read().strip()

    if synopsis_response_original != synopsis_response:
        print("Using new user edited synopsis:")
        print("```")
        print(synopsis_response)
        print("```")
    else:
        print("Synopsis unchanged.")

    # ------------------------------------------------------------------------------
    # Write chapters
    # ------------------------------------------------------------------------------
    start = time.time()
    p(f"Starting to write {len(synopsis_chapters)} chapters...")

    chapter_responses = []
    all_chapter_total_tokens = []
    pbar = tqdm(range(0, len(synopsis_chapters)))
    for i in pbar:
        chapter_number = i + 1
        pbar.set_description(f"Writing chapter {chapter_number}")
        p(f"Writing {chapter_number=}")

        is_first_chapter = chapter_number == 1
        total_chapters = len(synopsis_chapters)
        previous_chapter_text = "" if is_first_chapter else chapter_responses[chapter_number - 2]

        chapter_system = prompts.get_system("expand-chapter-first", llm_config)
        chapter_first_prompt_format = prompts.get_prompt("expand-chapter-first", llm_config)
        chapter_next_prompt_format = prompts.get_prompt("expand-chapter-next", llm_config)

        chapter_prompt = (
            chapter_first_prompt_format.format(
                synopsis_response=synopsis_response,
                total_chapters=total_chapters,
                book_description=book_description,
                book_characters=book_characters,
            )
            if is_first_chapter
            else chapter_next_prompt_format.format(
                previous_chapter_number=chapter_number - 1,
                previous_chapter_text=previous_chapter_text,
                synopsis_response=synopsis_response,
                chapter_number=chapter_number,
                total_chapters=total_chapters,
                book_description=book_description,
                book_characters=book_characters,
            )
        )

        chapter_response, chapter_total_tokens = llm.make_call(chapter_system, chapter_prompt, llm_config)
        all_chapter_total_tokens.append(chapter_total_tokens)
        chapter_response = chapter_response.replace(
            "```", ""  # TODO: HACK: investigate, can be introduced if in prompt template
        )
        chapter_responses.append(chapter_response)
        with open(output_folder / f"chapter_{chapter_number}.txt", "w") as f:
            f.write(f"{chapter_total_tokens=}")
            f.write(f"{len(all_chapter_total_tokens)=}")
            f.write(f"{sum(all_chapter_total_tokens)=}")
            f.write(f"{chapter_number=}")
            f.write("\n---\nchapter_prompt=\n")
            f.write(chapter_prompt)
            f.write("\n---\nchapter_response=\n")
            f.write(chapter_response)

    # ------------------------------------------------------------------------------
    # Construct whole book and write to file
    # ------------------------------------------------------------------------------
    total_process_tokens = synopsis_total_tokens + sum(all_chapter_total_tokens)
    rough_gpt4_8k_price_estimate = utils.gpt4_8k_price_estimate(total_process_tokens)
    rough_gpt35_4k_price_estimate = utils.gpt35_4k_price_estimate(total_process_tokens)

    whole_book = f"# {synopsis_title}"
    whole_book += "\n\n---\n\n"
    whole_book += "## Synopsis\n\n" + synopsis_response
    whole_book += "\n\n---\n\n"
    whole_book += "## Full Text"
    for chapter_response in chapter_responses:
        whole_book += "\n\n" + chapter_response
    whole_book += "\n\n---\n\n"
    whole_book += "## Technical Details\n\n" + str(safe_llm_config) + "\n\n"
    whole_book += f"Total process tokens: {total_process_tokens:,}\n\n"
    whole_book += f"Rough GPT4 8k price: ${rough_gpt4_8k_price_estimate:.2f} (estimated, [check your usage!](https://platform.openai.com/usage))\n\n"
    whole_book += f"Rough GPT3.5 4k price: ${rough_gpt35_4k_price_estimate:.3f} (estimated, [check your usage!](https://platform.openai.com/usage))\n\n"
    whole_book += f"Synopsis tokens: {synopsis_total_tokens:,}\n\n"
    whole_book += f"Sum of all chapter tokens: {sum(all_chapter_total_tokens):,}\n\n"
    whole_book += f"Average chapter tokens: {sum(all_chapter_total_tokens)/len(all_chapter_total_tokens):,.1f}\n\n"
    whole_book += f"Min chapter tokens: {min(all_chapter_total_tokens):,}\n\n"
    whole_book += f"Max chapter tokens: {max(all_chapter_total_tokens):,}\n\n"
    whole_book += f"Individual chapter tokens: {all_chapter_total_tokens}\n\n"
    whole_book += "\n\n---\n\n"
    whole_book += (
        "Plot written by a human, full text expanded by [GPTAuthor](https://github.com/dylanhogg/gptauthor).\n"
    )

    whole_book = replace_chapter_text(whole_book)
    with open(output_folder / "_whole_book.md", "w") as f:
        f.write(whole_book)

    whole_book_html = markdown.markdown(whole_book)
    html_style = 'max-width:1000px; margin-left:auto; margin-right:auto; border:1px solid #ddd; padding:25px; font-family:"Georgia","Times New Roman",serif; font-size: larger;'
    whole_book_html = f"<html><body><div style='{html_style}'>" + whole_book_html + "</div></body></html>"
    with open(output_folder / "_whole_book.html", "w") as f:
        f.write(whole_book_html)

    with open(output_folder / "__finished.txt", "w") as f:
        f.write(f"Finished writing book: {synopsis_title}.\n\n")
        f.write(str(safe_llm_config))

    took = time.time() - start

    # ------------------------------------------------------------------------------
    # Final output
    # ------------------------------------------------------------------------------
    p(f"\n{str(safe_llm_config)}")
    p(f"\nFinished writing book: {synopsis_title}")
    p(f"\nOutput written to '{output_folder}'")
    p(
        f"\n{took=:.2f}s, {total_process_tokens=:,}, ${rough_gpt4_8k_price_estimate=:.2f}, ${rough_gpt35_4k_price_estimate=:.2f} (estimated cost, check your usage!)"
    )

    if llm_config.allow_user_input:
        while True:
            user_input = input("Press 'Y' to open the html book in a browser, or Enter/Return to finish: ")
            if user_input.lower() == "y":
                abs_file = (output_folder / "_whole_book.html").resolve()
                webbrowser.open(f"file://{abs_file}")
            elif user_input.lower() == "":
                return
            else:
                print("Invalid input. Please try again.")

    return output_folder



import importlib.resources
from pathlib import Path

from loguru import logger
from omegaconf import DictConfig, OmegaConf

from gptauthor.library import consts
from gptauthor.library.classes import AppUsageException


def _get_common(key: str, llm_config: DictConfig):
    yaml_file = get_yaml_file(llm_config)
    conf_file = OmegaConf.load(yaml_file)
    if key not in conf_file:
        raise Exception(f"{key} not in conf_file")
    return conf_file[key]


def _get_conf(prompt_type: str, llm_config) -> (str, dict):
    yaml_file = get_yaml_file(llm_config)
    conf_file = OmegaConf.load(yaml_file)
    if prompt_type not in conf_file:
        valid_prompt_types = sorted(
            [k for k in conf_file.keys() if not k.startswith("_") and not k.startswith("unit_test")]
        )
        raise AppUsageException(f"Entity type '{prompt_type}' not supported. Try one of these: {valid_prompt_types}")
    conf = conf_file[prompt_type]
    assert "system" in conf, f"Prompt type '{prompt_type}' is missing a 'system' key"
    assert "prompt" in conf, f"Prompt type '{prompt_type}' is missing a 'prompt' key"
    return conf


def get_yaml_file(llm_config: DictConfig):
    local_yaml_file = Path(llm_config.story_file)
    if local_yaml_file.is_file():
        logger.info(f"Using local yaml file: {local_yaml_file}")
        return local_yaml_file

    resources_yaml_file = Path(importlib.resources.files(consts.package_name).joinpath(llm_config.story_file))
    if resources_yaml_file.is_file():
        logger.info(f"Using resources yaml file: {resources_yaml_file}")
        return resources_yaml_file

    raise Exception(
        f"Could not find yaml file: {llm_config.story_file} either locally or in the resources folder. "
        "See here for an example: https://github.com/dylanhogg/gptauthor/blob/main/gptauthor/prompts-openai-drama.yaml"
    )


def get_prompt(prompt_type: str, llm_config: DictConfig):
    return _get_conf(prompt_type, llm_config).prompt


def get_system(prompt_type: str, llm_config: DictConfig):
    return _get_conf(prompt_type, llm_config).system


def get_book_description(llm_config: DictConfig):
    return _get_common("common-book-description", llm_config)


def get_book_characters(llm_config: DictConfig):
    return _get_common("common-book-characters", llm_config)



from typing import Optional

import typer
from loguru import logger
from omegaconf import OmegaConf
from rich import print
from typing_extensions import Annotated

from .library import consts, engine, env, log
from .library.classes import AppUsageException

typer_app = typer.Typer()


def version_callback(value: bool):
    if value:
        print(f"{consts.package_name} version: {consts.version}")
        print("Please visit https://github.com/dylanhogg/gptauthor for more info.")
        raise typer.Exit()


@typer_app.command()
def run(
    story: Annotated[str, typer.Option(help="The name of the yaml file defining the story and prompts")],
    llm_model: Annotated[str, typer.Option(help="The model name")] = consts.default_llm_model,
    llm_temperature: Annotated[
        float, typer.Option(help="LLM temperature value (0 to 2, OpenAI default is 1)")
    ] = consts.default_llm_temperature,
    llm_top_p: Annotated[
        float, typer.Option(help="LLM top_p probability value (0 to 2, OpenAI default is 1)")
    ] = consts.default_llm_top_p,
    llm_use_localhost: Annotated[
        int, typer.Option(help="LLM use localhost:8081 instead of openai")
    ] = consts.default_llm_use_localhost,
    total_chapters: Annotated[int, typer.Option(help="Total chapters to write")] = consts.default_write_total_chapters,
    allow_user_input: Annotated[bool, typer.Option(help="Allow command line user input")] = True,
    version: Annotated[
        Optional[bool],
        typer.Option("--version", help=f"Display {consts.package_name} version", callback=version_callback),
    ] = None,
) -> None:
    """
    gptauthor entry point
    """

    try:
        log.configure()
        example_usage = f"Example usage: [bold green]{consts.package_name} --story prompts-openai-drama --total-chapters 3 --llm-model gpt-3.5-turbo --llm-temperature 0.1 --llm-top-p 1.0[/bold green]"

        llm_api_key = env.get("OPENAI_API_KEY", "")
        if not llm_use_localhost and not llm_api_key:
            raise AppUsageException(
                "Expected an environment variable 'OPENAI_API_KEY' to be set to use OpenAI API."
                "\nSee the OpenAI docs for more info: https://platform.openai.com/docs/quickstart/step-2-setup-your-api-key"
                "\nAlternatively you can use the '--llm_use_localhost 1' argument to use a local LLM server."
            )

        story_file = f"{story}.yaml"
        llm_config = OmegaConf.create(
            {
                "version": consts.version,
                "api_key": llm_api_key,
                "model": llm_model,
                "temperature": llm_temperature,
                "top_p": llm_top_p,
                "total_chapters": total_chapters,
                "use_localhost": llm_use_localhost,
                "localhost_sleep": int(env.get("LLM_USE_LOCALHOST_SLEEP", 0)),
                "default_output_folder": consts.default_output_folder,
                "story_file": story_file,
                "allow_user_input": allow_user_input,
            }
        )

        engine.do_writing(llm_config)

    except AppUsageException as ex:
        print(example_usage)
        print(f"[bold red]{str(ex)}[/bold red]")
        print("")
        print(f"For more information, try '{consts.package_name} --help'.")
        logger.exception(ex)

    except typer.Exit as ex:
        if ex.exit_code == 0:
            print()
            print(
                "[bold green]Goodbye and thanks for using gptauthor! Please visit https://github.com/dylanhogg/gptauthor for more info.[/bold green]"
            )
            return
        print(example_usage)
        print(f"[bold red]Unexpected error code: {str(ex)}[/bold red]")
        print("")
        print(f"For more information, try '{consts.package_name} --help'.")
        logger.exception(ex)

    except Exception as ex:
        print(example_usage)
        print(f"[bold red]Unexpected exception: {str(ex)}[/bold red]")
        print("")
        print(f"For more information, try '{consts.package_name} --help'.")
        logger.exception(ex)



import sys

from loguru import logger

from . import env


def configure(remove_existing=True, logfile="./log/app.log"):
    if remove_existing:
        logger.remove()
    logger.add(sys.stderr, level=env.get("LOG_STDERR_LEVEL", "ERROR"))
    logger.add(
        logfile,
        level=env.get("LOG_FILE_LEVEL", "WARNING"),
        rotation=env.get("LOG_FILE_ROTATION", "00:00"),
    )
